/*
 *   kgoodstuff - a button bar a la FvwmButtons for KDE
 *   Copyright (C) 1998  Juergen Vigna
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>

#include <qtooltip.h>
#include <qstring.h>
#include <qfiledlg.h>
#include <qregexp.h>
#include <qmsgbox.h>
#include <qfile.h>
#include <qtstream.h>
#include <qbitmap.h>
#include <qwindefs.h>

#include <kapp.h>
#include <kkeyconf.h>
#include <kwm.h>
#include <drag.h>
#include <kurl.h>

#include "kgoodstuff.h"

#define Inherited KGoodStuffData
/* Global Variables */

extern bool isKdelnkFile(const char* filename);

static int calcMaxWidth(QMyButton *button, QFont *font, int maxw)
{
    int
        w1,w2 = 0;

    w1 = QFontMetrics(*font).width(button->text());
    if (button->pixmap())
        w2 = button->pixmap()->width();
    w2 = (w2 > w1)? w2:w1;
    w2 += 4;
    return ((maxw > w2)? maxw:w2);
}

static int calcMaxHeight(QMyButton *button, QFont *font, int maxh)
{
    int
        height = 0;
    
    if (button->pixmap())
        height = button->pixmap()->height() + 2;
    if (button->text() && strlen(button->text()))
        height += QFontMetrics(*font).height() + 2;
    else
        height += 2;
    return ((maxh > height)? maxh:height);
}

void execute(const char* cmd)
{
  KShellProcess proc;
  proc << cmd;
  proc.start(KShellProcess::DontCare);
}

KGoodStuff::KGoodStuff(QWidget* parent, const char* name, const char *num,
                       bool childFlag, bool closeChildFlag,
                       KGoodStuff *master)
	:Inherited( parent, name )
{
    int
        i;
    QString
        cpixmap, cname, ctip, ccmd,
        path,
        tmpstr;
    QPoint
        pt;

    isChild = childFlag;
    closeChild = closeChildFlag;
    masterKGS = master;
    pt.setX(x());
    pt.setY(y());
    recreate(parent, WStyle_Customize | WStyle_NoBorder | WStyle_Tool ,pt);
//    recreate(parent, WStyle_Customize | WStyle_DialogBorder,pt);
//    recreate(parent, WStyle_NoBorder | WStyle_Tool ,pt);
    font = new QFont("helvetica", 12, 50, 0);
    CHECK_PTR(font);
    popup_menu = new QPopupMenu;
    CHECK_PTR( popup_menu);
    popup_menu->insertItem(i18n("&Help"),ID_MENU_HELP);
    popup_menu->insertItem(i18n("About"),ID_MENU_ABOUT);
    popup_menu->insertSeparator();
    popup_menu->insertItem(i18n("New"),ID_MENU_NEW);
    popup_menu->insertItem(i18n("Copy"),ID_MENU_COPY);
    popup_menu->insertItem(i18n("Paste"),ID_MENU_PASTE);
    popup_menu->insertItem(i18n("Cut"),ID_MENU_CUT);
    popup_menu->insertSeparator();
    popup_menu->insertItem(i18n("Button Options"),ID_MENU_BOPT);
    popup_menu->insertItem(i18n("Global Options"),ID_MENU_GOPT);
    popup_menu->insertItem(i18n("Save Options"),ID_MENU_SAVE);
    popup_menu->insertSeparator();
    if (isChild)
        popup_menu->insertItem(i18n("&Close"),ID_MENU_CLOSE);
    else
        popup_menu->insertItem(i18n("&Quit"),ID_MENU_QUIT);

    aConfigName = KApplication::localconfigdir() + "/kgoodstuffrc";
    KSimpleConfig
        config(aConfigName);
    debugFlag = getenv("DEBUG");

    font->setStyleHint( (QFont::StyleHint)5 );
    font->setCharSet( (QFont::CharSet)0 );

    kconfgroup = "kgoodstuff";
    kconfglob = "global";
    if (num) {
        kconfgroup += num;
        kconfglob += num;
    }
    config.setGroup(kconfglob);
    iconPath = config.readEntry("IconPath");
    maxXY = config.readEntry("MaxXY");
    minXY = config.readEntry("MinXY");
    posX = config.readEntry("PosX");
    posY = config.readEntry("PosY");
    hOrient = config.readNumEntry("Orientation");
    eRaise = config.readNumEntry("EnterRaise");
    lLower = config.readNumEntry("LeaveLower");
    nColumns = config.readNumEntry("RowColumns");
    prefglob = new PreferencesGlob(0,"Global Options");
    CHECK_PTR(prefglob);
    connect(prefglob, SIGNAL(okButtonPressed()), SLOT(prefGlobOK()));
    prefbutton = new PreferencesButton(0,"Button Options");
    CHECK_PTR(prefbutton);
    prefglob->setPrefs(iconPath, maxXY, minXY, posX, posY,hOrient,nColumns,
                       eRaise, lLower);
    iconPath = prefglob->getIconPath();
    maxXY = prefglob->getMaxXY();
    minXY = prefglob->getMinXY();
    posX = prefglob->getPosX();
    posY = prefglob->getPosY();
    nColumns = prefglob->getNColumns();
    config.writeEntry("TmpIconPath",iconPath.data()); // remove $$ form ENV
    config.sync();
    il = new KIconLoader((KConfig*)&config,kconfglob,"TmpIconPath");
    CHECK_PTR(il);
    PreferencesButton
        *pb = new PreferencesButton(il, 0, "Tab Dialog");
    delete prefbutton;
    prefbutton = pb;
    connect(prefbutton, SIGNAL(okButtonPressed()), SLOT(prefButtonOK()));
    config.setGroup(kconfgroup);
    nButtons = config.readNumEntry("NButtons");
    if (nButtons > 0) {
        buttons = new MyButtonStruct*[nButtons+1];
        CHECK_PTR(buttons);
        for(i=0;i<nButtons;i++) {
            tmpstr.sprintf("Button%d",i+1);
            cpixmap = tmpstr + ".pixmap";
            cname = tmpstr + ".name";
            ctip = tmpstr + ".desc";
            ccmd = tmpstr + ".command";
            buttons[i] = new MyButtonStruct;
            CHECK_PTR(buttons[i]);
            buttons[i]->button = new QMyButton(buttonGroup,tmpstr);
            CHECK_PTR(buttons[i]->button);
            buttons[i]->dz = new KDNDDropZone(buttons[i]->button, DndURL);
            CHECK_PTR(buttons[i]->dz);
            connect(buttons[i]->dz, SIGNAL(dropAction(KDNDDropZone *)),
                    this, SLOT(slotDropEvent(KDNDDropZone *)));
            buttons[i]->name = config.readEntry(cname);
            buttons[i]->button->setText(buttons[i]->name);
            buttons[i]->pixmap = config.readEntry(cpixmap);
            buttons[i]->button->setPixmap(loadIcon(buttons[i]->pixmap,false,
                                                   buttons[i]->button));
            buttons[i]->desc = config.readEntry(ctip);
            if (!buttons[i]->desc.isEmpty())
                QToolTip::add(buttons[i]->button, buttons[i]->desc);
            buttons[i]->cmd = config.readEntry(ccmd);
            buttons[i]->swallowName = config.readEntry(tmpstr+".swallowName");
            buttons[i]->swallowCmd=config.readEntry(tmpstr+".swallowCommand");
            buttons[i]->swallowed = 0;
            buttons[i]->button->setFont(*font);
            buttons[i]->childKGSFlag = config.readNumEntry(tmpstr+".childFlag",false);
            buttons[i]->button->setToggle(buttons[i]->childKGSFlag);
            buttons[i]->childCloseOpt = config.readNumEntry(tmpstr+".childCloseOpt",false);
            buttons[i]->useRows = config.readNumEntry(tmpstr+".useRows",1);
            buttons[i]->useCols = config.readNumEntry(tmpstr+".useCols",1);
            buttons[i]->childKGS = 0;
        }
        buttons[i] = 0;
    } else {
        buttons = 0;
    }
    copyButton = new MyButtonStruct;
    copyButton->button = new QMyButton(0,"CopyButton");
    copyButton->button->hide();
    emptyButton = new MyButtonStruct*[2];
    emptyButton[0] = new MyButtonStruct;
    CHECK_PTR(emptyButton[0]);
    emptyButton[1] = 0;
    emptyButton[0]->swallowed = 0;
    emptyButton[0]->button = new QMyButton(this,"Empty_Button");
    CHECK_PTR(emptyButton[0]->button);
    emptyButton[0]->button->setGeometry(0,0,60,30);
    emptyButton[0]->button->installEventFilter( this );
    emptyButton[0]->childKGSFlag = false;
    emptyButton[0]->childCloseOpt = false;
    emptyButton[0]->button->setToggle(false);
    emptyButton[0]->useRows = emptyButton[0]->useCols = 1;
    emptyButton[0]->childKGS = 0;
    noPopupMenu = FALSE;
    setButtonsGeometry();
    launchSwallowApplications();
    timer_raise = new QTimer(this);
    connect(timer_raise, SIGNAL(timeout()),this,SLOT(raise()));
    timer_lower = new QTimer(this);
    connect(timer_lower, SIGNAL(timeout()),this,SLOT(lower()));
}

KGoodStuff::~KGoodStuff()
{
    KSimpleConfig
        config(aConfigName);
    config.setGroup(kconfglob);
    config.deleteEntry("TmpIconPath",false);
    for (int i=0;i < nButtons; i++) {
        if (buttons[i]->swallowed) {
//            fprintf(stderr,"CLOSE %p\n",this);
            KWM::close(buttons[i]->swallowed);
            XFlush(qt_xdisplay());
        } else if (buttons[i]->childKGS) {
            delete buttons[i]->childKGS;
            buttons[i]->childKGS = 0;
            buttons[i]->button->setButtonOn(false);
        }
    }
}

void KGoodStuff::setButtonsGeometry()
{
    int
        maxh = 0, maxw = 0,
        actRC = 0, maxRC = 0,
        w,wx,hy,
        i,j,k,r,c,xp,yp;

    if (copyButton->name.isEmpty() &&
        copyButton->pixmap.isEmpty() &&
        copyButton->cmd.isEmpty() &&
        copyButton->desc.isEmpty() &&
        copyButton->swallowName.isEmpty())
        popup_menu->setItemEnabled(ID_MENU_PASTE,false);
    else
        popup_menu->setItemEnabled(ID_MENU_PASTE,true);
    if (!buttons || nButtons <= 0) {
        buttonGroup->insert(emptyButton[0]->button);
        buttonGroup->setGeometry(0,0,60,30);
        setFixedSize(60,30);
        emptyButton[0]->button->show();
        // disable the Menu items
        popup_menu->setItemEnabled(ID_MENU_CUT,false);
        popup_menu->setItemEnabled(ID_MENU_COPY,false);
        popup_menu->setItemEnabled(ID_MENU_BOPT,false);
        popup_menu->setItemEnabled(ID_MENU_SAVE,false);
        if (!posX.isEmpty() && !posY.isEmpty()) {
            int
                x = posX.toInt(),
                y = posY.toInt();
            if ((x < 0) || (posX.left(1) == "-"))
                x = kapp->desktop()->width() + x - width();
            if ((y < 0) || (posY.left(1) == "-"))
                y = kapp->desktop()->height() + y - height();
            setGeometry(x,y,60,30);
        } else if (isChild && masterKGS) {
            int
                x,y;
            bool
                or;
        
            if (masterKGS->getGeometryData(this,&x,&y,&or)) {
                if (or) {
                    y -= height();
                    if (y < height())
                        y += height() + masterKGS->height();
                    if (x+width() > kapp->desktop()->width())
                        x -= ((x+width()) - kapp->desktop()->width());
                } else {
                    x -= width();
                    if (x < width())
                        x += width() + masterKGS->width();
                    if (y+height() > kapp->desktop()->height())
                        y -= ((y+height()) - kapp->desktop()->height());
                }
                setGeometry(x,y,60,30);
            }
        }
        return;
    }
    emptyButton[0]->button->hide();
    buttonGroup->remove(emptyButton[0]->button);
    // enable the Menu items
    popup_menu->setItemEnabled(ID_MENU_CUT,true);
    popup_menu->setItemEnabled(ID_MENU_COPY,true);
    popup_menu->setItemEnabled(ID_MENU_BOPT,true);
    popup_menu->setItemEnabled(ID_MENU_SAVE,true);
    // calculate the really maximum number of rows
    for(i=0;i<nButtons;i++) {
        if (hOrient)
            maxRC += buttons[i]->useCols;
        else
            maxRC += buttons[i]->useRows;
    }
    // allocate bool table for maximum number of rows buttons
    bool
        fcst[maxRC][nColumns];
    for(i=0;i<maxRC;i++)
        for(j=0;j<nColumns;j++)
            fcst[i][j] = false;
    // First calculate new values for maxh and maxw
    for(k=0,i=0;i<nButtons;k++) {
        for(j=0;(j < nColumns) && (i < nButtons);j++,i++) {
            if (fcst[k][j]) {
                if (debugFlag)
                    fprintf(stderr,"SKIP:%d:%d\n",k,j);
                i--;
                continue;
            }
            buttonGroup->remove(buttons[i]->button);
            maxh = calcMaxHeight(buttons[i]->button,font,maxh);
            maxw = calcMaxWidth(buttons[i]->button,font,maxw);
            if (hOrient) {
                for(c=j;(c<nColumns) && ((c-j)<buttons[i]->useRows);c++)
                    for(r=k;(r-k)<buttons[i]->useCols;r++)
                        fcst[r][c] = true;
            } else {
                for(c=j;(c<nColumns) && ((c-j)<buttons[i]->useCols);c++)
                    for(r=k;(r-k)<buttons[i]->useRows;r++)
                        fcst[r][c] = true;
            }
            if (debugFlag) {
                for(c=0;c<nColumns;c++) {
                    for(r=0;r<maxRC;r++) {
                        if (fcst[r][c])
                            fprintf(stderr,"x");
                        else
                            fprintf(stderr,"o");
                    }
                    fprintf(stderr,"\n");
                }
            }
        }
    }
    // now calculate the max row needed
    for(c=0;c<nColumns;c++) {
        for(r=0;r<maxRC;r++) {
            if (!fcst[r][c])
                break;
        }
        if (actRC < r)
            actRC = r;
    }
    w = maxh > maxw ? maxh:maxw;
    if (!minXY.isEmpty())
        if (w < minXY.toInt())
            w = minXY.toInt();
    if (!maxXY.isEmpty())
        if (w > maxXY.toInt())
            w = maxXY.toInt() + 4;
    if (w <= 4)
        w = 30;
    hy = wx = w;
    if (hOrient) {
        hy = hy * nColumns + (2 * (nColumns+1));
        wx = wx * actRC + (2 * (actRC+1));
    } else {
        wx = wx * nColumns + (2 * (nColumns+1));
        hy = hy * actRC + (2 * (actRC+1));
    }
    buttonGroup->setGeometry(0,0,wx,hy);
    setFixedSize(wx,hy);
    if (!posX.isEmpty() && !posY.isEmpty()) {
        int
            x = posX.toInt(),
            y = posY.toInt();
        if ((x < 0) || (posX.left(1) == "-"))
            x = kapp->desktop()->width() + x - width();
        if ((y < 0) || (posY.left(1) == "-"))
            y = kapp->desktop()->height() + y - height();
        setGeometry(x,y,wx,hy);
    } else if (isChild && masterKGS) {
        int
            x,y;
        bool
            or;
        
        if (masterKGS->getGeometryData(this,&x,&y,&or)) {
            if (or) {
                y -= height();
                if (y < 0) // height())
                    y += height() + masterKGS->height();
                if (x+width() > kapp->desktop()->width())
                    x -= ((x+width()) - kapp->desktop()->width());
            } else {
                x -= width();
                if (x < 0) // width())
                    x += width() + masterKGS->width();
                if (y+height() > kapp->desktop()->height())
                    y -= ((y+height()) - kapp->desktop()->height());
            }
            setGeometry(x,y,wx,hy);
        }
    }
    xp = 2;
    yp = 2;
    // redo this for the new cycle as otherwise I would skip from the begining
    for(i=0;i<maxRC;i++)
        for(j=0;j<nColumns;j++)
            fcst[i][j] = false;
    for(i=0,k=0;i<nButtons;k++) {
        for(j=0;(j < nColumns) && (i < nButtons);j++,i++) {
            while((j < nColumns) && fcst[k][j]) {
                j++;
                if (hOrient)
                    yp += w + 2;
                else
                    xp += w + 2;
            }
            if (j >= nColumns)
                break;
            buttonGroup->insert(buttons[i]->button,i);
            buttons[i]->button->installEventFilter( this );
            if (hOrient) {
                // check for block in y-directon
                for(r=j+1;
                    (r<nColumns)&&((r-j) < buttons[i]->useRows)&&!fcst[k][r];
                    r++)
                    ;
                wx = (w * buttons[i]->useCols) + (2 * (buttons[i]->useCols-1));
                r = r-j;
                hy = (w * r) + (2 * (r-1));
                for(r=buttons[i]->useCols-1;r>=0;r--) {
                    if (buttons[i]->useRows+j > nColumns)
                        c = j - nColumns + 1;
                    else
                        c = buttons[i]->useRows-1;
                    for(;c>=0;c--) {
                        fcst[k+r][j+c] = true;
                    }
                }
                j += (buttons[i]->useRows-1);
            } else {
                // check for block in x-directon
                for(c=j+1;
                    (c<nColumns)&&((c-j) < buttons[i]->useCols)&&!fcst[k][c];
                    c++)
                    ;
                c = c-j;
                wx = (w * c) + (2 * (c-1));
                hy = (w * buttons[i]->useRows) + (2 * (buttons[i]->useRows-1));
                for(r=buttons[i]->useRows-1;r>=0;r--) {
                    if (buttons[i]->useCols+j > nColumns)
                        c = j - nColumns + 1;
                    else
                        c = buttons[i]->useCols-1;
                    for(;c>=0;c--) {
                        fcst[k+r][j+c] = true;
                    }
                }
                j += (buttons[i]->useCols-1);
            }
            if (xp+wx > width())
                wx = width() - xp - 2;
            if (yp+hy > height())
                hy = height() - yp - 2;
            buttons[i]->button->setGeometry(xp, yp, wx, hy);
            if (buttons[i]->childKGSFlag)
                buttons[i]->button->setPixmap(loadIcon(buttons[i]->pixmap,true,
                                                       buttons[i]->button));
            buttons[i]->button->show();
            if (hOrient)
                yp += w*buttons[i]->useRows + 2;
            else
                xp += (w*buttons[i]->useCols) + 2;
        }
        if (hOrient) {
            yp = 2;
            xp += w + 2;
        } else {
            xp = 2;
            yp += w + 2;
        }
    }
}

void KGoodStuff::handleButtons(int id)
{
    QString
        name, cmd;
    QMyButton
        *actButton;
    int
        bid;

    actButton = (QMyButton *)buttonGroup->find(id);
    for(bid=0;buttons && buttons[bid] && (buttons[bid]->button!=actButton);bid++)
        ;
    name = ((QString)actButton->name()) + ".command";
    if (actButton && actButton->last_button == MidButton) {
        return;
    } else if (!buttons || (actButton->last_button == RightButton)) {
        if (noPopupMenu) {
            noPopupMenu = FALSE;
            return;
        }
        handle_popup(popup_menu,actButton);
    	if (buttons && buttons[bid])
            buttons[bid]->button->setMouseIsOn(false);
        return;
    }
    if (buttons[bid]->childKGSFlag) {
        if (buttons[bid]->childKGS) {
            if (buttons[bid]->childKGS->isVisible()) {
                buttons[bid]->childKGS->hide();
                buttons[bid]->childKGS->removeChilds();
//                delete buttons[bid]->childKGS;
//                buttons[bid]->childKGS = 0;
                buttons[bid]->button->setButtonOn(false);
            } else {
                buttons[bid]->childKGS->setButtonsGeometry();
                buttons[bid]->childKGS->show();
                buttons[bid]->button->setButtonOn(true);
            }
        } else if (buttons[bid]->name.isEmpty()) {
            QString
                str = i18n("No Name defined!!!");
            QMessageBox::warning(this,"KGoodStuff",str);
        } else {
            buttons[bid]->childKGS =
                new KGoodStuff(0,0,buttons[bid]->name,true,
                               buttons[bid]->childCloseOpt,this);
            if (buttons[bid]->childKGS->iconPath.isEmpty())
                buttons[bid]->childKGS->setIconPath(iconPath);
            buttons[bid]->childKGS->setButtonsGeometry();
            buttons[bid]->childKGS->show();
            connect(buttons[bid]->childKGS, SIGNAL(closeButtonBar(KGoodStuff *)),
                    this, SLOT(closeKGS(KGoodStuff *)));
            buttons[bid]->button->setButtonOn(true);
        }
        return;
    }
    if (buttons[bid]->cmd.isEmpty())
        return;
    executeCommand(buttons[bid]->cmd);
    if (closeChild)
        emit closeButtonBar(this);
}

void KGoodStuff::handleMenu(int)
{
}

void KGoodStuff::handle_popup(QPopupMenu *menu, QWidget *button)
{
    int
        i,bid,
        v,xp,yp;
    QWidget
        *d = kapp->desktop();
    MyButtonStruct
        **but = 0;

    for(bid=0;buttons && (buttons[bid]->button!=button);bid++)
        ;
    menu->move(-1000,-1000);
    menu->show();
    xp = button->mapToGlobal(QPoint(0,0)).x() + button->width();
    yp = button->mapToGlobal(QPoint(0,0)).y();
    if ((xp + menu->width()) > d->width())
        xp -= (button->width() + menu->width());
    if ((yp + menu->height()) > d->height())
        yp -= (menu->height() - button->height());
    menu->move(xp,yp);
    switch((v=menu->exec())) {
      case ID_MENU_NEW:
      {
          QString
              tmpstr;

          nButtons++;
          tmpstr.sprintf("Button%d",nButtons);
          but = new MyButtonStruct*[nButtons+1];
          CHECK_PTR(but);
          for(i=0;i<bid;i++)
              but[i] = buttons[i];
          but[bid] = new MyButtonStruct;
          CHECK_PTR(but[bid]);
          but[bid]->swallowed = 0;
          but[bid]->button = new QMyButton(buttonGroup,tmpstr);
          CHECK_PTR(but[bid]->button);
          but[bid]->button->setFont(*font);
          but[bid]->dz = new KDNDDropZone(but[bid]->button, DndURL);
          CHECK_PTR(but[bid]->dz);
          connect(but[bid]->dz, SIGNAL(dropAction(KDNDDropZone *)),
                  this, SLOT(slotDropEvent(KDNDDropZone *)));
          but[bid]->childKGSFlag = false;
          but[bid]->childCloseOpt = false;
          but[bid]->button->setToggle(false);
          but[bid]->childKGS = 0;
          but[bid]->useCols = but[bid]->useRows = 1;
          for(i=bid+1; i < nButtons; i++)
              but[i] = buttons[i-1];
          but[i] = 0;
          if (buttons)
              delete buttons;
          buttons = but;
          setButtonsGeometry();
      }
      break;
      case ID_MENU_CUT:
          if (!buttons)
              return;
          nButtons--;
          if (nButtons <= 0) {
              buttonGroup->remove(buttons[bid]->button);
              if (buttons[bid]->button->pixmap())
                  copyButton->button->setPixmap(
                      *buttons[bid]->button->pixmap());
              copyButton->button->setText(buttons[bid]->name);
              copyButton->name = buttons[bid]->name;
              copyButton->pixmap = buttons[bid]->pixmap;
              copyButton->cmd = buttons[bid]->cmd;
              copyButton->desc = buttons[bid]->desc;
              copyButton->swallowName = buttons[bid]->swallowName;
              copyButton->swallowCmd = buttons[bid]->swallowCmd;
              copyButton->childKGSFlag = buttons[bid]->childKGSFlag;
              copyButton->childCloseOpt = buttons[bid]->childCloseOpt;
              copyButton->button->setToggle(copyButton->childKGSFlag);
              copyButton->childKGS = 0;
              copyButton->useRows = buttons[bid]->useRows;
              copyButton->useCols = buttons[bid]->useCols;
              delete buttons[bid]->button;
              delete buttons[bid]->dz;
              delete buttons[bid];
              delete buttons;
              buttons = 0;
              setButtonsGeometry();
              return;
          }
          but = new MyButtonStruct*[nButtons+1];
          CHECK_PTR(but);
          for(i=0;i<bid;i++)
              but[i] = buttons[i];
          buttonGroup->remove(buttons[bid]->button);
          if (buttons[bid]->button->pixmap())
              copyButton->button->setPixmap(*buttons[bid]->button->pixmap());
          copyButton->button->setText(buttons[bid]->name);
          copyButton->name = buttons[bid]->name;
          copyButton->pixmap = buttons[bid]->pixmap;
          copyButton->cmd = buttons[bid]->cmd;
          copyButton->desc = buttons[bid]->desc;
          copyButton->swallowName = buttons[bid]->swallowName;
          copyButton->swallowCmd = buttons[bid]->swallowCmd;
          copyButton->childKGSFlag = buttons[bid]->childKGSFlag;
          copyButton->childCloseOpt = buttons[bid]->childCloseOpt;
          copyButton->button->setToggle(copyButton->childKGSFlag);
          copyButton->childKGS = 0;
          copyButton->useRows = buttons[bid]->useRows;
          copyButton->useCols = buttons[bid]->useCols;
          delete buttons[bid]->button;
          delete buttons[bid]->dz;
          delete buttons[bid];
          for(i=bid+1; i <= nButtons; i++)
              but[i-1] = buttons[i];
          but[i-1] = 0;
          delete buttons;
          buttons = but;
          setButtonsGeometry();
          break;
      case ID_MENU_COPY:
          if (!buttons)
              return;
          if (buttons[bid]->button->pixmap())
              copyButton->button->setPixmap(*buttons[bid]->button->pixmap());
          copyButton->button->setText(buttons[bid]->name);
          copyButton->name = buttons[bid]->name;
          copyButton->pixmap = buttons[bid]->pixmap;
          copyButton->cmd = buttons[bid]->cmd;
          copyButton->desc = buttons[bid]->desc;
          copyButton->swallowName = buttons[bid]->swallowName;
          copyButton->swallowCmd = buttons[bid]->swallowCmd;
          copyButton->childKGSFlag = buttons[bid]->childKGSFlag;
          copyButton->childCloseOpt = buttons[bid]->childCloseOpt;
          copyButton->button->setToggle(copyButton->childKGSFlag);
          copyButton->childKGS = 0;
          copyButton->useRows = buttons[bid]->useRows;
          copyButton->useCols = buttons[bid]->useCols;
          if (copyButton->name.isEmpty() &&
              copyButton->pixmap.isEmpty() &&
              copyButton->cmd.isEmpty() &&
              copyButton->desc.isEmpty() &&
              copyButton->swallowName.isEmpty())
              popup_menu->setItemEnabled(ID_MENU_PASTE,false);
          else
              popup_menu->setItemEnabled(ID_MENU_PASTE,true);
          break;
      case ID_MENU_PASTE:
          if (copyButton->name.isEmpty() &&
              copyButton->pixmap.isEmpty() &&
              copyButton->cmd.isEmpty() &&
              copyButton->desc.isEmpty() &&
              copyButton->swallowName.isEmpty())
              return;
      {
          QString
              name,cmd,
              tmpstr;

          nButtons++;
          tmpstr.sprintf("Button%d",nButtons);
          but = new MyButtonStruct*[nButtons+1];
          CHECK_PTR(but);
          for(i=0;i<bid;i++)
              but[i] = buttons[i];
          but[bid] = new MyButtonStruct;
          CHECK_PTR(but[bid]);
          but[bid]->swallowed = 0;
          but[bid]->button = new QMyButton(buttonGroup,tmpstr);
          CHECK_PTR(but[bid]->button);
          but[bid]->button->setFont(*font);
          but[bid]->dz = new KDNDDropZone(but[bid]->button, DndURL);
          CHECK_PTR(but[bid]->dz);
          connect(but[bid]->dz, SIGNAL(dropAction(KDNDDropZone *)),
                  this, SLOT(slotDropEvent(KDNDDropZone *)));
          but[bid]->pixmap = copyButton->pixmap;
          if (copyButton->button->pixmap())
              but[bid]->button->setPixmap(*(copyButton->button->pixmap()));
          but[bid]->name = copyButton->name;
          but[bid]->button->setText(copyButton->name);
          but[bid]->cmd = copyButton->cmd;
          but[bid]->desc = copyButton->desc;
          but[bid]->swallowName = copyButton->swallowName;
          but[bid]->swallowCmd = copyButton->swallowCmd;
          but[bid]->childKGSFlag = copyButton->childKGSFlag;
          but[bid]->childCloseOpt = copyButton->childCloseOpt;
          but[bid]->button->setToggle(copyButton->childKGSFlag);
          but[bid]->childCloseOpt = copyButton->childCloseOpt;
          but[bid]->childKGS = 0;
          but[bid]->useRows = copyButton->useRows;
          but[bid]->useCols = copyButton->useCols;
          if (!but[bid]->desc.isEmpty())
              QToolTip::add(but[bid]->button, but[bid]->desc);
          for(i=bid+1; i < nButtons; i++)
              but[i] = buttons[i-1];
          but[i] = 0;
          delete buttons;
          buttons = but;
          setButtonsGeometry();
      }
          break;
      case ID_MENU_GOPT:
          prefglob->setPrefs(iconPath, maxXY, minXY, posX, posY,hOrient,
                             nColumns, eRaise, lLower);
          prefglob->show();
          break;
      case ID_MENU_BOPT:
          if (!buttons)
              return;
          prefbutton->setPrefs(buttons[bid]->name, buttons[bid]->pixmap,
                               buttons[bid]->cmd, buttons[bid]->desc,
                               buttons[bid]->swallowName,
                               buttons[bid]->swallowCmd,
                               buttons[bid]->childKGSFlag,
                               buttons[bid]->childCloseOpt,
                               buttons[bid]->useRows,
                               buttons[bid]->useCols);
          prefbutton->data = (void *)buttons[bid];
          prefbutton->show();
          break;
      case ID_MENU_SAVE:
          saveOptions();
          break;
      case ID_MENU_HELP: {
          kapp->invokeHTMLHelp("", "");
          break;
      }
      case ID_MENU_ABOUT: {
          QMessageBox::about(0,
              i18n("About"),i18n("KGoodStuff 0.6.1 \n\nCopyright 1998\n"
                                 "by Jrgen Vigna\njug@sad.it"));
          break;
      }
      case ID_MENU_CLOSE:
          removeChilds();
          emit closeButtonBar(this);
          break;
      case ID_MENU_QUIT:
          for (i=0;i < nButtons; i++) {
              if (buttons[i]->swallowed) {
                  KWM::close(buttons[i]->swallowed);
                  XFlush(qt_xdisplay());
              } else if (buttons[i]->childKGS) {
                  delete buttons[i]->childKGS;
                  buttons[i]->childKGS = 0;
                  buttons[i]->button->setButtonOn(false);
              }
          }
          kapp->quit();
          exit(0);
          break;
      default:
          break;
    }
}

void KGoodStuff::saveOptions()
{
    saveGlobalOptions();
    saveButtonOptions();
    for(int i=0;i<nButtons;i++) {
        if (buttons[i]->childKGS)
            buttons[i]->childKGS->saveOptions();
    }
}

void KGoodStuff::saveGlobalOptions()
{
    QString
        tstr;
    KSimpleConfig
        config(aConfigName);
    
    config.setGroup(kconfglob);
    tstr = iconPath.data();
    tstr.replace(QRegExp("\\$"),"$$");
    config.writeEntry("IconPath",tstr);
    if (maxXY.isEmpty())
        config.deleteEntry("MaxXY",false);
    else
        config.writeEntry("MaxXY",maxXY.toInt());
    if (minXY.isEmpty())
        config.deleteEntry("MinXY",false);
    else
        config.writeEntry("MinXY",minXY.toInt());
    if (posX.isEmpty())
        config.deleteEntry("PosX",false);
    else
        config.writeEntry("PosX",posX);
    if (posY.isEmpty())
        config.deleteEntry("PosY",false);
    else
        config.writeEntry("PosY",posY);
    config.writeEntry("Orientation",hOrient);
    config.writeEntry("EnterRaise",eRaise);
    config.writeEntry("LeaveLower",lLower);
    config.writeEntry("RowColumns",nColumns);
}

void KGoodStuff::saveButtonOptions()
{
    QString
        bstr,tstr,
        cpixmap, cname, ctip, ccmd;
    int
        i;

    KSimpleConfig
        config(aConfigName);
    
    config.deleteGroup(kconfgroup);
    config.setGroup(kconfgroup);
    config.writeEntry("NButtons",nButtons);
    for(i=0;i<nButtons;i++) {
        bstr.sprintf("Button%d",i+1);
        cpixmap = bstr + ".pixmap";
        cname = bstr + ".name";
        ctip = bstr + ".desc";
        ccmd = bstr + ".command";
        config.writeEntry(cname,buttons[i]->name);
        config.writeEntry(cpixmap,buttons[i]->pixmap);
        tstr = buttons[i]->cmd.data();
        tstr.replace(QRegExp("\\$"),"$$");
        config.writeEntry(ccmd, tstr);
        tstr = buttons[i]->desc.data();
        tstr.replace(QRegExp("\\$"),"$$");
        config.writeEntry(ctip,tstr);
        tstr = buttons[i]->swallowName.data();
        tstr.replace(QRegExp("\\$"),"$$");
        config.writeEntry(bstr+".swallowName",tstr);
        tstr = buttons[i]->swallowCmd.data();
        tstr.replace(QRegExp("\\$"),"$$");
        config.writeEntry(bstr+".swallowCommand",tstr);
        config.writeEntry(bstr+".childFlag",buttons[i]->childKGSFlag);
        config.writeEntry(bstr+".childCloseOpt",buttons[i]->childCloseOpt);
        config.writeEntry(bstr+".useRows",buttons[i]->useRows);
        config.writeEntry(bstr+".useCols",buttons[i]->useCols);
    }
}

void KGoodStuff::ReadStderr(KProcess *, char *buffer, int len)
{
    fprintf(stderr,"%.*s",len,buffer);
}

void KGoodStuff::ReadStdout(KProcess *, char *buffer, int len)
{
    fprintf(stdout,"%.*s",len,buffer);
}

QPixmap KGoodStuff::loadIcon(const char *pname, bool flag=false,
                             QMyButton *button=0)
{
    QPixmap
        px;
    if (!pname || !strlen(pname)) {
        px = il->loadIcon("unknown.xpm");
        if (flag)
            arrow_on_pixmap(&px, button);
        return px;
    }
    if (maxXY.isEmpty()) {
        px = il->loadIcon(pname);
        if (flag)
            arrow_on_pixmap(&px, button);
        return px;
    }
    int
        w,h,d,
        nw = maxXY.toInt();

    px = il->loadIcon(pname);
    w = px.width();
    h = px.height();
    if (w == h) {
        w = nw;
        h = nw;
    } else if (w > h) {
        d = (w*100) / h;
        w = nw;
        h = (nw * 10000) / d;
        h += 500;
        h /= 100;
    } else {
        d = (h*100) / w;
        h = nw;
        w = (nw * 10000) / d;
        w += 500;
        w /= 100;
    }
    px = il->loadIcon(pname,w,h);
    if (flag)
        arrow_on_pixmap(&px, button);
    return px;
}

#define MOVE_NOTHING 0
#define MOVE_BUTTONBAR 1
#define MOVE_BUTTON  2

bool KGoodStuff::eventFilter(QObject *ob, QEvent *ev)
{
    static QWidget
        *moving_button = 0;
    static QPoint
        mb_offset;
    static int
        xpos, ypos, moveWhat = MOVE_NOTHING;

    switch (ev->type()) {
      case Event_MouseButtonDblClick:
      case Event_MouseButtonPress: {
          QMouseEvent* mev = (QMouseEvent*)ev;
          if (mev->button() == MidButton) {
              timer_raise->stop();
              if (ob->isWidgetType() && !((QWidget*)ob)->isPopup()){
                  moveWhat = MOVE_BUTTONBAR;
                  moving_button = (QWidget*)ob;
                  moving_button->setCursor(sizeAllCursor);
                  mb_offset = moving_button->mapToGlobal(mev->pos());
                  for(int i=0;buttons && buttons[i];i++) {
                      if (buttons[i]->childKGS)
                          buttons[i]->childKGS->myMove(mb_offset,true);
                  }
                  mb_offset = mb_offset - pos();
                  moving_button->setCursor(sizeAllCursor);
                  xpos = x();
                  ypos = y();
                  raise();
                  XGrabPointer( qt_xdisplay(), moving_button->winId(), FALSE,
                                ButtonPressMask | ButtonReleaseMask |
                                PointerMotionMask | EnterWindowMask |
                                LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                                None, None, CurrentTime );
              }
          } else if (mev->button() == RightButton) {
              timer_raise->stop();
              if (ob->isWidgetType() && !((QWidget*)ob)->isPopup()){
                  moveWhat = MOVE_BUTTON;
                  moving_button = (QWidget*)ob;
                  moving_button->setCursor(sizeAllCursor);
                  moving_button->raise();
                  mb_offset =
                      moving_button->mapToGlobal(mev->pos());
                  mb_offset =
                      moving_button->mapFromGlobal(mb_offset);
                  moving_button->setCursor(sizeAllCursor);
                  xpos = moving_button->x();
                  ypos = moving_button->y();
                  raise();
                  XGrabPointer( qt_xdisplay(), moving_button->winId(), FALSE,
                                ButtonPressMask | ButtonReleaseMask |
                                PointerMotionMask | EnterWindowMask |
                                LeaveWindowMask, GrabModeAsync, GrabModeAsync,
                                None, None, CurrentTime );
              }
          }
      }
      break;
      case Event_MouseButtonRelease: {
          if (moving_button) {
              if (moveWhat == MOVE_BUTTONBAR) {
                  if (xpos != x()) {
                      if (x() < (kapp->desktop()->width()/2))
                          posX.sprintf("%d",x());
                      else
                          posX.sprintf("-%d",-(x() - kapp->desktop()->width() +
                                               width()));
                  }
                  if (ypos != y()) {
                      if (y() < (kapp->desktop()->height()/2))
                          posY.sprintf("%d",y());
                      else
                          posY.sprintf("-%d",-(y() - kapp->desktop()->height()+
                                               height()));
                  }
              } else if ((ypos != moving_button->y()) ||
                         (xpos != moving_button->x())) {
                  if ((abs(ypos - moving_button->y()) > 5) ||
                      (abs(xpos - moving_button->x()) > 5))
                      noPopupMenu = TRUE;
                  moveButtonPosition((QMyButton*)moving_button,xpos,ypos);
              }
              moving_button->setCursor(arrowCursor);
              XUngrabPointer(qt_xdisplay(),0);
              moving_button = 0;
              moveWhat = MOVE_NOTHING;
          }
      }
      break;
      case Event_MouseMove: {
          if (moving_button) {
              if (moveWhat == MOVE_BUTTONBAR) {
                  QMouseEvent
                      *mev = (QMouseEvent*)ev;
                  for(int i=0;buttons && buttons[i];i++) {
                      if (buttons[i]->childKGS)
                          buttons[i]->childKGS->myMove(
                              moving_button->mapToGlobal(mev->pos()),false);
                  }
                  move(moving_button->mapToGlobal(mev->pos()) - mb_offset);
              } else {
                  int x = mapFromGlobal(QCursor::pos()).x()
                      - mb_offset.x();
                  int y = mapFromGlobal(QCursor::pos()).y()
                      - mb_offset.y();
                  if (hOrient) {
                      if (x < 0)
                          x=0;
                      if (x + moving_button->width() > width()) 
                          x = width() - moving_button->width ();
                      if (nColumns <= 1)
                          y = moving_button->y();
                      else {
                          if (y < 0)
                              y = 0;
                          if (y > height()-moving_button->height())
                              y = height()-moving_button->height();
                      }
                  } else {
                      if (y < 0)
                          y = 0;
                      if (y + moving_button->height() > height()) 
                          y = height() - moving_button->height ();
                      if (nColumns <= 1)
                          x = moving_button->x();
                      else {
                          if (x < 0)
                              x = 0;
                          if (x > width()-moving_button->width())
                              x = width()-moving_button->width();
                      }
                  }
                  moving_button->move(x, y);
              }
              return TRUE;
          }
      }
      break;
    }
    return FALSE;
}

void KGoodStuff::prefGlobOK()
{
    int i;
    
    if (prefglob->getIconPath() != iconPath) {
        setIconPath(prefglob->getIconPath());
    }
    if (prefglob->getMaxXY() != maxXY) {
        maxXY = prefglob->getMaxXY();
        if (buttons) {
            for(i=0;i<nButtons;i++) {
                buttons[i]->button->setPixmap(
                    loadIcon(buttons[i]->pixmap,false,buttons[i]->button));
            }
        }
    }
    minXY = prefglob->getMinXY();
    posX = prefglob->getPosX();
    posY = prefglob->getPosY();
    nColumns = prefglob->getNColumns();
    hOrient = prefglob->getHOrient();
    eRaise = prefglob->getERaise();
    lLower = prefglob->getLLower();
    updateSwallowButtons();
    setButtonsGeometry();
    saveOptions();
}

void KGoodStuff::prefButtonOK()
{
    MyButtonStruct
        *button = (MyButtonStruct *)prefbutton->data;
    bool
        updFlag=false;

    if (!button)
        return;
    if (button->childKGSFlag != prefbutton->getButtonChildKGS()) {
        if (button->childKGS) {
            button->childKGS->hide();
            delete button->childKGS;
            button->childKGS = 0;
            button->button->setButtonOn(false);
        }
        if (button->childKGSFlag)
            button->button->setPixmap(loadIcon(button->pixmap,false,
                                               button->button));
        button->childKGSFlag = prefbutton->getButtonChildKGS();
        button->button->setToggle(button->childKGSFlag);
    }
    button->childCloseOpt = prefbutton->getButtonChildCloseOpt();
    if (button->childKGS)
        button->childKGS->closeChild = button->childCloseOpt;
    if (prefbutton->getButtonName() != button->name) {
        QString
            str = prefbutton->getButtonName();

        if (button->childKGSFlag) {
            QString
                kg = "kgoodstuff";
            int
                ar=0;
            {
                KSimpleConfig
                    config(aConfigName);
                config.setGroup(kg+str);
                if (config.hasKey("NButtons")) {
                    ar = QMessageBox::information
                        (0,i18n("Cannot change name"),
                         i18n("There already exists a group with that name.\n"
                              "If  this  is an old  group which is not used\n"
                              "anymore you may delete it and  substitute it\n"
                              "with the new one.\n\n"
                              "ATTENTION: ARE YOU REALLY SURE?\n"),
                         i18n("Delete"),i18n("Cancel"),0, 0,2);
                }
            }
            if (ar == 0) {
                // remove the group already there with the new name
                // and save the actual one with this new name
                bool
                    delkgs = !button->childKGS;
                if (delkgs) {
                    button->childKGS = new KGoodStuff(
                        0,0,button->name,true,button->childCloseOpt,this);
                    if (button->childKGS->iconPath.isEmpty())
                        button->childKGS->setIconPath(iconPath);
                }
                {
                    KSimpleConfig
                        config(aConfigName);
                    kg = "kgoodstuff" + button->name;
                    config.setGroup(kg);
                    config.deleteGroup(kg);
                    kg = "global" + button->name;
                    config.setGroup(kg);
                    config.deleteGroup(kg);
                    config.sync();
                }
                button->childKGS->changeName(str);
                button->name = str;
                button->button->setText(str);
                saveOptions();
                if (delkgs) {
                    delete button->childKGS;
                    button->childKGS = 0;
                }
            }
        } else {
            button->name = prefbutton->getButtonName();
            button->button->setText(button->name);
        }
    }
    if (prefbutton->getButtonPixmap() != button->pixmap) {
        button->pixmap = prefbutton->getButtonPixmap();
        button->button->setPixmap(loadIcon(button->pixmap,false,
                                           button->button));
    }
    if (prefbutton->getButtonDesc() != button->desc) {
        button->desc = prefbutton->getButtonDesc();
        if (!button->desc.isEmpty())
            QToolTip::add(button->button, button->desc);
        else
            QToolTip::remove(button->button);
    }
    if (prefbutton->getButtonCommand() != button->cmd) {
        button->cmd = prefbutton->getButtonCommand();
        updFlag = true;
    }
    if (prefbutton->getButtonSwallowName() != button->swallowName) {
        button->swallowName = prefbutton->getButtonSwallowName();
        updFlag = true;
    }
    if (prefbutton->getButtonSwallowCmd() != button->swallowCmd) {
        button->swallowCmd = prefbutton->getButtonSwallowCmd();
        updFlag = true;
    }
    if (prefbutton->getButtonUseRows() != button->useRows) {
        button->useRows = prefbutton->getButtonUseRows();
        updFlag = true;
    }
    if (prefbutton->getButtonUseCols() != button->useCols) {
        button->useCols = prefbutton->getButtonUseCols();
        updFlag = true;
    }
    if (updFlag) {
        if (button->swallowed) {
            KWM::close(button->swallowed);
            button->swallowed = 0;
            XFlush(qt_xdisplay());
        }
        if (!button->swallowName.isEmpty() && !button->swallowCmd.isEmpty()) {
            KWM::doNotManage(button->swallowName);
            execute(button->swallowCmd);
        }
    }
    setButtonsGeometry();
    saveOptions();
}

void KGoodStuff::slotDropEvent(KDNDDropZone * dz)
{
    QStrList &list = dz->getURLList();
    QString
        path,
        s = list.first();
    QMyButton
        *actButton;
    int
        bid;

    path = KURL(s).path();
    if (!KURL(s).isLocalFile() || !QFileInfo(path).exists()) {
        QMessageBox::warning(0, i18n("Error cannot find file:"),
                             path, i18n("Ok"));
        return;
    }
    actButton = (QMyButton *)dz->getWidget();
    for(bid=0;buttons && buttons[bid] && (buttons[bid]->button!=actButton);bid++)
        ;
    if (!isKdelnkFile(path)) {
        if (buttons[bid]->cmd.isEmpty()) {
            buttons[bid]->cmd = "xterm -e sh -c \"" + path;
            buttons[bid]->cmd += ";echo ";
            buttons[bid]->cmd += i18n("'Press ENTER to continue'");
            buttons[bid]->cmd += ";read x\"";
            if (buttons[bid]->pixmap.isEmpty())
                buttons[bid]->pixmap = "unknown.xpm";
            buttons[bid]->button->setPixmap(loadIcon(buttons[bid]->pixmap,
                                                     false,
                                                     buttons[bid]->button));
            return;
        }
        executeCommand(buttons[bid]->cmd,path);
        return;
    }
    KSimpleConfig kconfig(path, true);
    kconfig.setGroup("KDE Desktop Entry");
    buttons[bid]->name = kconfig.readEntry("Name");
    buttons[bid]->pixmap = kconfig.readEntry("Icon");
    buttons[bid]->cmd = kconfig.readEntry("Exec");
    buttons[bid]->desc = kconfig.readEntry("Comment");
    buttons[bid]->swallowName = kconfig.readEntry("SwallowTitle");
    buttons[bid]->swallowCmd = kconfig.readEntry("SwallowExec");
    if (!buttons[bid]->pixmap.isEmpty())
        buttons[bid]->button->setPixmap(loadIcon(buttons[bid]->pixmap,
                                                 false,
                                                 buttons[bid]->button));
    buttons[bid]->button->setText(buttons[bid]->name);
    if (!buttons[bid]->desc.isEmpty())
        QToolTip::add(buttons[bid]->button, buttons[bid]->desc);
    setButtonsGeometry();
}

void KGoodStuff::closeKGS(KGoodStuff *kgs)
{
    int bid;
    
    for(bid=0;buttons && buttons[bid] && (buttons[bid]->childKGS!=kgs);bid++)
        ;
    
    buttons[bid]->childKGS->hide();
//    delete buttons[bid]->childKGS;
//    buttons[bid]->childKGS = 0;
    buttons[bid]->button->setButtonOn(false);
}

void KGoodStuff::executeCommand(const char *ecmd, const char *fname=0)
{
    KShellProcess
        *proc = new KShellProcess;
    QString
        cmd = ecmd;

    cmd.replace(QRegExp("%f"),fname);
    *proc << cmd.data();
    connect(proc, SIGNAL(receivedStderr(KProcess *,char*,int)),this, 
            SLOT(ReadStderr(KProcess*,char*,int)));
    connect(proc, SIGNAL(receivedStdout(KProcess *,char*,int)),this, 
            SLOT(ReadStdout(KProcess*,char*,int)));
    bool result = proc->start(KProcess::DontCare , KProcess::All);
    if(!result) {
        QString str;
        str.sprintf(i18n("Cannot start a new program\nfork() failed."));
        QMessageBox::critical(this,"KGoodStuff",str.data());
        if (proc) {
            if(proc->isRunning())
                proc->kill();
            delete proc;
        }
        return;
    }
}

void KGoodStuff::moveButtonPosition(QMyButton *button, int oldx, int oldy)
{
    int
        bid,obid,
        ly=button->y(),
        lx=button->x();
    MyButtonStruct
        *tmpbutton;
                  
    for(bid=0;buttons && (buttons[bid]->button!=button);bid++)
        ;
    tmpbutton = buttons[bid];
    if (hOrient) {
        obid = bid;
        while(bid && (lx < buttons[bid-1]->button->x()))
            bid--;
        while(((bid+1) < nButtons) &&
              (lx > buttons[bid+1]->button->x()))
            bid++;
        if (nColumns > 1) {
            if (button != buttons[bid]->button)
                lx = buttons[bid]->button->x();
            else if (((bid+1)<nButtons) && (((bid+1) % nColumns) > 0))
                lx = buttons[bid+1]->button->x();
            else if ((bid+1)<nButtons)
                lx = buttons[bid-1]->button->x();
            else
                lx = oldx;
            button->move(oldx,oldy);
            // move y to the top
            while(bid && (lx == buttons[bid-1]->button->x()))
                bid--;
            // move y to the bottom if needed
            while(((bid+1) < nButtons) &&
                  (buttons[bid]->button->x()==buttons[bid+1]->button->x()) &&
                  (ly > buttons[bid]->button->y()))
                bid++;
        }
    } else {
        obid = bid;
        while(bid && (ly < buttons[bid-1]->button->y()))
            bid--;
        while(((bid+1) < nButtons) &&
              (ly > buttons[bid+1]->button->y()))
            bid++;
        if (nColumns > 1) {
            if (button != buttons[bid]->button)
                ly = buttons[bid]->button->y();
            else if (((bid+1)<nButtons) && (((bid+1) % nColumns) > 0))
                ly = buttons[bid+1]->button->y();
            else if ((bid+1)<nButtons)
                ly = buttons[bid-1]->button->y();
            else
                ly = oldy;
            button->move(oldx,oldy);
            // move x to the left
            while(bid && (ly == buttons[bid-1]->button->y()))
                bid--;
            // move x to the right if needed
            while(((bid+1) < nButtons) &&
                  (buttons[bid]->button->y()==buttons[bid+1]->button->y()) &&
                  (lx > buttons[bid]->button->x())) {
                bid++;
            }
        }
    }
    if (obid < bid) {
        for(;obid < bid;obid++)
            buttons[obid] = buttons[obid+1];
    } else if (obid > bid) {
        for(;obid > bid;obid--)
            buttons[obid] = buttons[obid-1];
    }
    buttons[bid] = tmpbutton;
    setButtonsGeometry();
}

void KGoodStuff::arrow_on_pixmap(QPixmap* pm, QMyButton *button)
{
    static char * link_xpm[] = {
        /* width height num_colors chars_per_pixel */
        "12 11 2 1",
        /* colors */
        "       c black",
        ".      c white",
        /* pixels */
        "            ",
        " .......... ",
        " .... ..... ",
        " ...  ..... ",
        " ..       . ",
        " .        . ",
        " ..       . ",
        " ...  ..... ",
        " .... ..... ",
        " .......... ",
        "            "
    };
    static QPixmap
        link_pixmap((const char **)link_xpm);
    QPainter
        paint;
    QPixmap
        tmp_pixmap;
    int
        w,h;

    tmp_pixmap = *pm;
    if (!button)
        button = buttons[0]->button;
    pm->resize(button->width()-4,
               button->height()-4-QFontMetrics(*font).height());
    w = pm->width();
    h = pm->height();
    pm->setMask(0);
    paint.begin(pm);
    paint.eraseRect(0,0,w,h);
    paint.drawPixmap(w/2 - tmp_pixmap.width()/2,
                     h/2 - tmp_pixmap.height()/2,tmp_pixmap);
    paint.drawPixmap(0,0,link_pixmap);
    paint.end();
    pm->setMask(pm->createHeuristicMask());
}

void KGoodStuff::setIconPath(const char *newIconPath)
{
    KSimpleConfig
        config(aConfigName);
    int i;

    iconPath = newIconPath;
    config.setGroup(kconfglob);
    config.writeEntry("TmpIconPath",iconPath);
    config.sync();
    KIconLoader
        *lil = new KIconLoader((KConfig*)&config,kconfglob,"TmpIconPath");
    CHECK_PTR(lil);
    PreferencesButton
        *pb = new PreferencesButton(lil, 0, "Tab Dialog");
    MyButtonStruct
        *button = (MyButtonStruct *)prefbutton->data;
    connect(pb, SIGNAL(okButtonPressed()), SLOT(prefButtonOK()));
    if (button) {
        pb->setPrefs(button->name,button->pixmap,button->cmd,button->desc,
                     button->swallowName, button->swallowCmd,
                     button->childKGSFlag,button->childCloseOpt,
                     button->useRows, button->useCols);
        pb->data = (void *)button;
        if (prefbutton->isVisible()) {
            prefbutton->hide();
            pb->setGeometry(prefbutton->x(),prefbutton->y(),
                            prefbutton->width(),prefbutton->height());
            delete prefbutton;
            delete il;
            il = lil;
            prefbutton = pb;
            pb->show();
        } else {
            delete prefbutton;
            delete il;
            il = lil;
            prefbutton = pb;
        }
    } else {
        delete prefbutton;
        delete il;
        il = lil;
        prefbutton = pb;
    }
    if (buttons) {
        for(i=0;i<nButtons;i++) {
            buttons[i]->button->setPixmap(
                loadIcon(buttons[i]->pixmap,false,buttons[i]->button));
        }
    }
}

bool KGoodStuff::getGeometryData(KGoodStuff *child, int *x, int *y, bool *or)
{
    int i;
    
    for(i=0;(i<nButtons) && (buttons[i]->childKGS != child);i++)
        ;
    if (i >= nButtons)
        return false;
    if (hOrient) {
        *y = this->y();
        *x = this->x() + buttons[i]->button->x();
    } else {
        *x = this->x();
        *y = this->y() + buttons[i]->button->y();
    }
    *or = hOrient;
    return true;
}

void KGoodStuff::windowAdd(Window w)
{
    { 
        // ignore transient windows
        Window trans = None;
        if (XGetTransientForHint(qt_xdisplay(), w, &trans)){
            if (trans != None && trans != qt_xrootwin())
                return;
        }
    }
    // swallowing?
    (void)localWindowAdd(w);
}

bool KGoodStuff::localWindowAdd(Window w)
{
    QString t = KWM::titleWithState(w);
//    fprintf(stderr,"In WindowAdd [%s]\n",t.data());
    int i,bi;
    bool found = false;
    QRegExp r;
    for (bi=0; bi<nButtons && !found; bi++){
        if (!buttons[bi]->swallowName.isEmpty() && !buttons[bi]->swallowed) {
            r = buttons[bi]->swallowName;
            found = r.match(t) != -1;
        }
    }
    if (found) {
        bi--;
        for (i=0; i<nButtons && buttons[i]->swallowed != w;i++);
        if (i == nButtons) {
            buttons[bi]->button->setText("");
            buttons[bi]->swallowed = w;
            KWM::prepareForSwallowing(w);
            XSetWindowBackground(qt_xdisplay(),w,
                                 buttons[bi]->button->backgroundColor().pixel());
            XReparentWindow(qt_xdisplay(), w,
                            buttons[bi]->button->winId(), 3, 3);
            XSelectInput(qt_xdisplay(), w, EnterWindowMask | LeaveWindowMask);
            XResizeWindow(qt_xdisplay(), w,
                          buttons[bi]->button->width()-6,
                          buttons[bi]->button->height()-6);
            XMapWindow(qt_xdisplay(),w);
            if (buttons[bi]->cmd.isEmpty()){
                XGrabButton(qt_xdisplay(),
                            Button2,
                            AnyModifier, w, True, 
                            ButtonPressMask,
                            GrabModeSync, GrabModeAsync, None, None);
                XGrabButton(qt_xdisplay(),
                            Button3,
                            AnyModifier, w, True, 
                            ButtonPressMask,
                            GrabModeSync, GrabModeAsync, None, None);
            } else {
                XGrabButton(qt_xdisplay(), AnyButton, AnyModifier, w, True, 
                            ButtonPressMask,
                            GrabModeSync, GrabModeAsync, None, None);
            }
            buttons[bi]->button->swallowed_window = w;
        }
    } else {
        for (bi=0; bi<nButtons && !found; bi++){
            if (buttons[bi]->childKGS)
                found = buttons[bi]->childKGS->localWindowAdd(w);
        }
    }
    return found;
}

void KGoodStuff::removeChilds()
{
    for (int i=0;i < nButtons; i++) {
        if (buttons[i]->childKGS) {
            delete buttons[i]->childKGS;
            buttons[i]->childKGS = 0;
            buttons[i]->button->setButtonOn(false);
        }
    }
}

void KGoodStuff::launchSwallowApplications()
{
    for (int i=0;i < nButtons; i++) {
        if (!buttons[i]->swallowCmd.isEmpty() && !buttons[i]->swallowed) {
            KWM::doNotManage(buttons[i]->swallowName);
            execute(buttons[i]->swallowCmd);
        }
    }
}

QWidget* KGoodStuff::parentOfSwallowed(Window w)
{
    int
        i;
    QWidget
        *bt=0;
  
    for (i=0; (i<nButtons) && (buttons[i]->swallowed != w) && !bt; i++) {
        if (buttons[i]->childKGS)
            bt=buttons[i]->childKGS->parentOfSwallowed(w);
    }
    if (bt)
        return bt;
    if (i<nButtons)
        return buttons[i]->button;
    return 0;
}

void KGoodStuff::updateSwallowButtons()
{
    for (int i=0;i < nButtons; i++) {
        if (buttons[i]->swallowed) {
            KWM::close(buttons[i]->swallowed);
            buttons[i]->swallowed = 0;
            XFlush(qt_xdisplay());
        }
    }
    launchSwallowApplications();
}

void KGoodStuff::enterEvent(QEvent *)
{
    timer_lower->stop();
    if (eRaise > 0) {
        timer_raise->start(eRaise,true);
        for(int i=0;buttons && buttons[i] && buttons[i]->button;++i) {
            if (buttons[i]->childKGS)
                buttons[i]->childKGS->enterEvent(0);
        }
    }
}

void KGoodStuff::leaveEvent(QEvent *)
{
    timer_raise->stop();
    if (lLower > 0) {
        timer_lower->start(lLower,true);
        for(int i=0;buttons && buttons[i] && buttons[i]->button;++i) {
            if (buttons[i]->childKGS)
                buttons[i]->childKGS->leaveEvent(0);
        }
    }
}

void KGoodStuff::changeName(const char *name)
{
    kconfgroup = "kgoodstuff";
    kconfglob = "global";
    if (name) {
        kconfgroup += name;
        kconfglob += name;
    }
}

void KGoodStuff::myMove(QPoint point, bool init)
{
    if (!posX.isEmpty() || !posY.isEmpty())
        return;
    if (init) {
        mymv_offset =  point - pos();
        return;
    }
    move(point - mymv_offset);
}
